@@ -166,7 +166,7 @@ class TestsExceptionFilter(_SQLAExceptionMatcher, test_base.BaseTestCase):
             yield
 
     def _run_test(self, dialect_name, statement, raises, expected,
-                  is_disconnect=False, params=()):
+                  is_disconnect=False, params=None):
         with self._fixture(dialect_name, raises, is_disconnect=is_disconnect):
             with self.engine.connect() as conn:
                 matched = self.assertRaises(
@@ -1006,6 +1006,12 @@ class TestDeadlock(TestsExceptionFilter):
             orig_exception_cls=self.InternalError
         )
 
+        self._run_deadlock_detect_test(
+            "mysql",
+            "(1213, 'Deadlock: wsrep aborted transaction')",
+            orig_exception_cls=self.OperationalError
+        )
+
     def test_mysql_pymysql_galera_deadlock(self):
         self._run_deadlock_detect_test(
             "mysql",
@@ -1546,7 +1552,8 @@ class TestDBConnectRetry(TestsExceptionFilter):
 
         with self._dbapi_fixture(dialect_name):
             with mock.patch.object(engine.dialect, "connect", cant_connect):
-                return engines._test_connection(engine, retries, .01)
+                return engines._test_connection(
+                    engine, retries, .01, _close=False)
 
     def test_connect_no_retries(self):
         conn = self._run_test(
